@@ -233,7 +233,7 @@ static gboolean   g_daemon_file_input_stream_close             (GInputStream
 								GCancellable         *cancellable,
 								GError              **error);
 static GFileInfo *g_daemon_file_input_stream_query_info        (GFileInputStream     *stream,
-								char                 *attributes,
+								const char           *attributes,
 								GCancellable         *cancellable,
 								GError              **error);
 static goffset    g_daemon_file_input_stream_tell              (GFileInputStream     *stream);
@@ -271,7 +271,7 @@ static gboolean   g_daemon_file_input_stream_close_finish      (GInputStream
 								GAsyncResult         *result,
 								GError              **error);
 static void       g_daemon_file_input_stream_query_info_async  (GFileInputStream     *stream,
-								char                 *attributes,
+								const char           *attributes,
 								int                   io_priority,
 								GCancellable         *cancellable,
 								GAsyncReadyCallback   callback,
@@ -735,7 +735,7 @@ iterate_read_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 				 current_len + len);
 	      io_op->io_buffer = file->input_buffer->str + current_len;
 	      io_op->io_size = len;
-	      io_op->io_allow_cancel = !op->sent_cancel;
+	      io_op->io_allow_cancel = file->input_buffer->len == 0 && !op->sent_cancel;
 	      return STATE_OP_READ;
 	    }
 
@@ -993,7 +993,7 @@ iterate_close_state_machine (GDaemonFileInputStream *file, IOOperationData *io_o
 				 current_len + len);
 	      io_op->io_buffer = file->input_buffer->str + current_len;
 	      io_op->io_size = len;
-	      io_op->io_allow_cancel = !op->sent_cancel;
+	      io_op->io_allow_cancel = file->input_buffer->len == 0 && !op->sent_cancel;
 	      return STATE_OP_READ;
 	    }
 
@@ -1257,7 +1257,7 @@ iterate_seek_state_machine (GDaemonFileInputStream *file, IOOperationData *io_op
 				 current_len + len);
 	      io_op->io_buffer = file->input_buffer->str + current_len;
 	      io_op->io_size = len;
-	      io_op->io_allow_cancel = !op->sent_cancel;
+	      io_op->io_allow_cancel = file->input_buffer->len == 0 && !op->sent_cancel;
 	      return STATE_OP_READ;
 	    }
 
@@ -1534,7 +1534,7 @@ iterate_query_state_machine (GDaemonFileInputStream *file,
 				 current_len + len);
 	      io_op->io_buffer = file->input_buffer->str + current_len;
 	      io_op->io_size = len;
-	      io_op->io_allow_cancel = !op->sent_cancel;
+	      io_op->io_allow_cancel = file->input_buffer->len == 0 && !op->sent_cancel;
 	      return STATE_OP_READ;
 	    }
 
@@ -1591,7 +1591,7 @@ iterate_query_state_machine (GDaemonFileInputStream *file,
 
 static GFileInfo *
 g_daemon_file_input_stream_query_info (GFileInputStream     *stream,
-				       char                 *attributes,
+				       const char           *attributes,
 				       GCancellable         *cancellable,
 				       GError              **error)
 {
@@ -1606,7 +1606,7 @@ g_daemon_file_input_stream_query_info (GFileInputStream     *stream,
   memset (&op, 0, sizeof (op));
   op.state = QUERY_STATE_INIT;
   if (attributes)
-    op.attributes = attributes;
+    op.attributes = (char *)attributes;
   else
     op.attributes = "";
   
@@ -1631,6 +1631,7 @@ typedef void (*AsyncIteratorDone) (GInputStream *stream,
 				   gpointer op_data,
 				   GAsyncReadyCallback callback,
 				   gpointer callback_data,
+                                   GCancellable *cancellable,
 				   GError *io_error);
 
 struct AsyncIterator {
@@ -1654,6 +1655,7 @@ async_iterator_done (AsyncIterator *iterator, GError *io_error)
 		     iterator->iterator_data,
 		     iterator->callback,
 		     iterator->callback_data,
+                     iterator->cancellable,
 		     io_error);
 
   g_free (iterator);
@@ -1829,6 +1831,7 @@ async_read_done (GInputStream *stream,
 		 gpointer op_data,
 		 GAsyncReadyCallback callback,
 		 gpointer user_data,
+                 GCancellable *cancellable,
 		 GError *io_error)
 {
   ReadOperation *op;
@@ -1859,7 +1862,7 @@ async_read_done (GInputStream *stream,
     g_simple_async_result_set_from_error (simple, error);
 
   /* Complete immediately, not in idle, since we're already in a mainloop callout */
-  g_simple_async_result_complete (simple);
+  _g_simple_async_result_complete_with_cancellable (simple, cancellable);
   g_object_unref (simple);
   
   if (op->ret_error)
@@ -1941,6 +1944,7 @@ async_close_done (GInputStream *stream,
 		  gpointer op_data,
 		  GAsyncReadyCallback callback,
 		  gpointer user_data,
+                  GCancellable *cancellable,
 		  GError *io_error)
 {
   GDaemonFileInputStream *file;
@@ -1948,7 +1952,6 @@ async_close_done (GInputStream *stream,
   CloseOperation *op;
   gboolean result;
   GError *error;
-  GCancellable *cancellable = NULL; /* TODO: get cancellable */
 
   file = G_DAEMON_FILE_INPUT_STREAM (stream);
   
@@ -1984,7 +1987,7 @@ async_close_done (GInputStream *stream,
     g_simple_async_result_set_from_error (simple, error);
 
   /* Complete immediately, not in idle, since we're already in a mainloop callout */
-  g_simple_async_result_complete (simple);
+  _g_simple_async_result_complete_with_cancellable (simple, cancellable);
   g_object_unref (simple);
   
   if (op->ret_error)
@@ -2029,6 +2032,7 @@ async_query_done (GInputStream *stream,
 		  gpointer op_data,
 		  GAsyncReadyCallback callback,
 		  gpointer user_data,
+                  GCancellable *cancellable,
 		  GError *io_error)
 {
   GDaemonFileInputStream *file;
@@ -2063,7 +2067,7 @@ async_query_done (GInputStream *stream,
 					       g_object_unref);
   
   /* Complete immediately, not in idle, since we're already in a mainloop callout */
-  g_simple_async_result_complete (simple);
+  _g_simple_async_result_complete_with_cancellable (simple, cancellable);
   g_object_unref (simple);
   
   if (op->ret_error)
@@ -2074,7 +2078,7 @@ async_query_done (GInputStream *stream,
 
 static void
 g_daemon_file_input_stream_query_info_async  (GFileInputStream     *stream,
-					      char                 *attributes,
+					      const char           *attributes,
 					      int                   io_priority,
 					      GCancellable         *cancellable,
 					      GAsyncReadyCallback   callback,
