@@ -33,6 +33,8 @@
 #include "gdaemonfile.h"
 #include "gvfsdaemonprotocol.h"
 #include "gdbusutils.h"
+#include "gmountsource.h"
+#include "gmountoperationdbus.h"
 
 /* Protects all fields of GDaemonMount that can change
    which at this point is just foreign_volume */
@@ -43,7 +45,6 @@ struct _GDaemonMount {
 
   GMountInfo *mount_info;
 
-  GVolume *foreign_volume;
   GVolumeMonitor *volume_monitor;
 };
 
@@ -60,9 +61,6 @@ g_daemon_mount_finalize (GObject *object)
   
   mount = G_DAEMON_MOUNT (object);
 
-  if (mount->foreign_volume != NULL)
-    g_object_unref (mount->foreign_volume);
-
   if (mount->volume_monitor != NULL)
     g_object_remove_weak_pointer (G_OBJECT (mount->volume_monitor), (gpointer) &(mount->volume_monitor));
 
@@ -112,7 +110,8 @@ g_daemon_mount_get_root (GMount *mount)
 {
   GDaemonMount *daemon_mount = G_DAEMON_MOUNT (mount);
 
-  return g_daemon_file_new (daemon_mount->mount_info->mount_spec, "/");
+  return g_daemon_file_new (daemon_mount->mount_info->mount_spec, 
+        daemon_mount->mount_info->mount_spec->mount_prefix);
 }
 
 static GIcon *
@@ -120,7 +119,7 @@ g_daemon_mount_get_icon (GMount *mount)
 {
   GDaemonMount *daemon_mount = G_DAEMON_MOUNT (mount);
 
-  return g_themed_icon_new_with_default_fallbacks (daemon_mount->mount_info->icon);
+  return g_object_ref (daemon_mount->mount_info->icon);
 }
 
 static char *
@@ -140,24 +139,13 @@ g_daemon_mount_get_uuid (GMount *mount)
 static GVolume *
 g_daemon_mount_get_volume (GMount *mount)
 {
-  GDaemonMount *daemon_mount = G_DAEMON_MOUNT (mount);
-  if (daemon_mount->foreign_volume != NULL)
-    return g_object_ref (daemon_mount->foreign_volume);
   return NULL;
 }
 
 static GDrive *
 g_daemon_mount_get_drive (GMount *mount)
 {
-  GDaemonMount *daemon_mount = G_DAEMON_MOUNT (mount);
-  GDrive *drive;
-
-  G_LOCK (daemon_mount);
-  drive = NULL;
-  if (daemon_mount->foreign_volume != NULL)
-    drive = g_volume_get_drive (daemon_mount->foreign_volume);
-  G_UNLOCK (daemon_mount);
-  return drive;
+  return NULL;
 }
 
 static gboolean
@@ -169,60 +157,7 @@ g_daemon_mount_can_unmount (GMount *mount)
 static gboolean
 g_daemon_mount_can_eject (GMount *mount)
 {
-  GDaemonMount *daemon_mount = G_DAEMON_MOUNT (mount);
-  gboolean res;
-  
-  G_LOCK (daemon_mount);
-  res = FALSE;
-  if (daemon_mount->foreign_volume != NULL)
-    res = g_volume_can_eject (daemon_mount->foreign_volume);
-  G_UNLOCK (daemon_mount);
-  
-  return res;
-}
-
-static void
-foreign_volume_removed (GVolume *volume, gpointer user_data)
-{
-  GDaemonMount *daemon_mount = G_DAEMON_MOUNT (user_data);
-
-  G_LOCK (daemon_mount);
-
-  g_object_ref (daemon_mount);
-  
-  if (daemon_mount->foreign_volume == volume)
-    {
-      g_object_unref (daemon_mount->foreign_volume);
-      daemon_mount->foreign_volume = NULL;
-    }
-
-  G_UNLOCK (daemon_mount);
-  
-  g_signal_emit_by_name (daemon_mount, "changed");
-  if (daemon_mount->volume_monitor != NULL)
-    g_signal_emit_by_name (daemon_mount->volume_monitor, "mount_changed", daemon_mount);
-  
-  g_object_unref (daemon_mount);
-}
-
-void
-g_daemon_mount_set_foreign_volume (GDaemonMount *mount, 
-                                   GVolume *foreign_volume)
-{
-  G_LOCK (daemon_mount);
-  
-  if (mount->foreign_volume != NULL)
-    g_object_unref (mount->foreign_volume);
-
-  if (foreign_volume != NULL)
-    {
-      mount->foreign_volume = foreign_volume;
-      g_signal_connect_object (foreign_volume, "removed", (GCallback) foreign_volume_removed, mount, 0);
-    }
-  else
-    mount->foreign_volume = NULL;
-
-  G_UNLOCK (daemon_mount);
+  return FALSE;
 }
 
 static void
@@ -241,17 +176,20 @@ unmount_reply (DBusMessage *reply,
 }
 
 static void
-g_daemon_mount_unmount (GMount *mount,
-			GMountUnmountFlags flags,
-			GCancellable *cancellable,
-			GAsyncReadyCallback callback,
-			gpointer         user_data)
+g_daemon_mount_unmount_with_operation (GMount *mount,
+                                       GMountUnmountFlags flags,
+                                       GMountOperation *mount_operation,
+                                       GCancellable *cancellable,
+                                       GAsyncReadyCallback callback,
+                                       gpointer         user_data)
 {
   GDaemonMount *daemon_mount = G_DAEMON_MOUNT (mount);
   DBusMessage *message;
   GMountInfo *mount_info;
   GSimpleAsyncResult *res;
   guint32 dbus_flags;
+  GMountSource *mount_source;
+  const char *dbus_id, *obj_path;
 
   mount_info = daemon_mount->mount_info;
 
@@ -261,18 +199,45 @@ g_daemon_mount_unmount (GMount *mount,
 				  G_VFS_DBUS_MOUNT_INTERFACE,
 				  G_VFS_DBUS_MOUNT_OP_UNMOUNT);
 
+  mount_source = g_mount_operation_dbus_wrap (mount_operation, _g_daemon_vfs_get_async_bus ());
+  dbus_id = g_mount_source_get_dbus_id (mount_source);
+  obj_path = g_mount_source_get_obj_path (mount_source);
+
   dbus_flags = flags;
-  _g_dbus_message_append_args (message, DBUS_TYPE_UINT32, &dbus_flags, 0);
+  _g_dbus_message_append_args (message,
+                               DBUS_TYPE_STRING, &dbus_id, DBUS_TYPE_OBJECT_PATH, &obj_path,
+                               DBUS_TYPE_UINT32, &dbus_flags,
+                               0);
   
   res = g_simple_async_result_new (G_OBJECT (mount),
 				   callback, user_data,
-				   g_daemon_mount_unmount);
+				   g_daemon_mount_unmount_with_operation);
   
   _g_vfs_daemon_call_async (message,
 			    unmount_reply, res,
 			    cancellable);
   
   dbus_message_unref (message);
+
+  g_object_unref (mount_source);
+}
+
+static gboolean
+g_daemon_mount_unmount_with_operation_finish (GMount *mount,
+                                              GAsyncResult *result,
+                                              GError **error)
+{
+  return TRUE;
+}
+
+static void
+g_daemon_mount_unmount (GMount *mount,
+			GMountUnmountFlags flags,
+			GCancellable *cancellable,
+			GAsyncReadyCallback callback,
+			gpointer         user_data)
+{
+  g_daemon_mount_unmount_with_operation (mount, flags, NULL, cancellable, callback, user_data);
 }
 
 static gboolean
@@ -280,81 +245,7 @@ g_daemon_mount_unmount_finish (GMount *mount,
 				GAsyncResult *result,
 				GError **error)
 {
-  return TRUE;
-}
-
-typedef struct {
-  GObject *object;
-  GAsyncReadyCallback callback;
-  gpointer user_data;
-} EjectWrapperOp;
-
-static void
-eject_wrapper_callback (GObject *source_object,
-                        GAsyncResult *res,
-                        gpointer user_data)
-{
-  EjectWrapperOp *data  = user_data;
-  data->callback (data->object, res, data->user_data);
-  g_free (data);
-}
-
-static void
-g_daemon_mount_eject (GMount              *mount,
-		      GMountUnmountFlags   flags,
-                      GCancellable        *cancellable,
-                      GAsyncReadyCallback  callback,
-                      gpointer             user_data)
-{
-  GDaemonMount *daemon_mount = G_DAEMON_MOUNT (mount);
-  GDrive *drive;
-
-  G_LOCK (daemon_mount);
-
-  drive = NULL;
-  if (daemon_mount->foreign_volume != NULL)
-    drive = g_volume_get_drive (G_VOLUME (daemon_mount->foreign_volume));
-      
-  G_UNLOCK (daemon_mount);
-  
-  if (drive != NULL)
-    {
-      EjectWrapperOp *data;
-      data = g_new0 (EjectWrapperOp, 1);
-      data->object = G_OBJECT (mount);
-      data->callback = callback;
-      data->user_data = user_data;
-      g_drive_eject (drive, flags, cancellable, eject_wrapper_callback, data);
-      g_object_unref (drive);
-    }
-}
-
-static gboolean
-g_daemon_mount_eject_finish (GMount        *mount,
-			     GAsyncResult  *result,
-			     GError       **error)
-{
-  GDaemonMount *daemon_mount = G_DAEMON_MOUNT (mount);
-  GDrive *drive;
-  gboolean res;
-
-  res = TRUE;
-
-  G_LOCK (daemon_mount);
-
-  drive = NULL;
-  if (daemon_mount->foreign_volume != NULL)
-    drive = g_volume_get_drive (G_VOLUME (daemon_mount->foreign_volume));
-
-  G_UNLOCK (daemon_mount);
-  
-  if (drive != NULL)
-    {
-      res = g_drive_eject_finish (drive, result, error);
-      g_object_unref (drive);
-    }
-  
-  return res;
+  return g_daemon_mount_unmount_with_operation_finish (mount, result, error);
 }
 
 static char **
@@ -389,7 +280,7 @@ g_daemon_mount_guess_content_type (GMount              *mount,
                                       callback,
                                       user_data,
                                       NULL);
-  g_simple_async_result_complete (simple);
+  g_simple_async_result_complete_in_idle (simple);
   g_object_unref (simple);
 }
 
@@ -414,8 +305,8 @@ g_daemon_mount_mount_iface_init (GMountIface *iface)
   iface->can_eject = g_daemon_mount_can_eject;
   iface->unmount = g_daemon_mount_unmount;
   iface->unmount_finish = g_daemon_mount_unmount_finish;
-  iface->eject = g_daemon_mount_eject;
-  iface->eject_finish = g_daemon_mount_eject_finish;
+  iface->unmount_with_operation = g_daemon_mount_unmount_with_operation;
+  iface->unmount_with_operation_finish = g_daemon_mount_unmount_with_operation_finish;
   iface->guess_content_type = g_daemon_mount_guess_content_type;
   iface->guess_content_type_finish = g_daemon_mount_guess_content_type_finish;
   iface->guess_content_type_sync = g_daemon_mount_guess_content_type_sync;
