@@ -26,7 +26,6 @@
 #include <string.h>
 #include <glib.h>
 #include <glib/gi18n.h>
-#include <glib/gurifuncs.h>
 #include <gio/gio.h>
 
 #include <avahi-client/client.h>
@@ -80,6 +79,12 @@ static struct {
           FALSE,
           "folder-remote-ssh"
         },
+	{
+	  "_smb._tcp",
+	  "smb",
+	  FALSE,
+	  "network-server"
+	},
 };
 
 static AvahiClient *global_client = NULL;
